use crate::{config::Config, util, util::CliStyle};
use addonscript::manifest::{
    Contributor,
    File,
    Manifest,
    Meta,
    Relation,
    Repository,
    RepositoryType,
};
use anyhow::Context;
use crossterm::style::Stylize;
use futures::{stream, StreamExt};
use indicatif::ProgressBar;
use reqwest::Client;
use serde::Serialize;
use std::{collections::HashMap, path::PathBuf, sync::Arc};
use tera::Tera;
use twitch::api::AddonInfoResponse;

const TEMPLATE: &str = include_str!("../../assets/modlist.html.tera");

pub async fn run(
    (config, mut manifest): (Config, Manifest),
    outfile: PathBuf,
) -> anyhow::Result<()> {
    let version = manifest
        .versions
        .pop()
        .context("Manifest has no versions!")?;
    let repos = util::repo_map(manifest.repositories);

    let mut tera = Tera::default();
    tera.add_raw_template("modlist", TEMPLATE)?;

    let http = Arc::new(Client::new());

    let len = version.relations.len();

    let mut futures = stream::iter(version.relations.into_iter())
        .map(|rel| get_meta(Arc::clone(&http), rel, &repos))
        .buffer_unordered(config.downloads.max_threads);

    let mut metas = vec![];
    let pb = ProgressBar::new(len as u64)
        .with_style(util::progress_style())
        .with_prefix("Resolving metadata");
    while let Some(meta) = futures.next().await {
        pb.inc(1);
        let meta = meta?;
        pb.println(format!(
            "{} {}",
            "Got meta for".green(),
            AsRef::<str>::as_ref(&meta.name).cyan().bold()
        ));
        metas.push(meta);
    }
    pb.finish();

    metas.sort_by_key(|m| m.name.to_ascii_lowercase());

    println!("{}", "Rendering modlist.".info());

    let rendered = tera.render(
        "modlist",
        &tera::Context::from_serialize(&ModListContent {
            metas,
            pack_meta: manifest.meta,
        })?,
    )?;

    println!("{}", "Writing file.".info());
    if let Some(parent) = outfile.parent() {
        tokio::fs::create_dir_all(parent).await?;
    }
    tokio::fs::write(outfile, rendered).await?;

    Ok(())
}

async fn get_meta(
    http: Arc<Client>,
    rel: Relation,
    repos: &HashMap<String, Repository>,
) -> anyhow::Result<Meta> {
    if let Some(meta) = rel.meta {
        return Ok(meta);
    }

    if rel.file.is_none() {
        return Ok(Meta {
            name: rel.id.clone(),
            contributors: vec![],
            description: rel.versions.map(|v| format!("version {}", v)),
            icon_url: None,
            website_url: None,
        });
    }

    let file = rel.file.unwrap();

    match file {
        File::Link { link, id, .. } => {
            Ok(Meta {
                name: if let Some(id) = id {
                    id
                } else {
                    util::link_file_name(&link)?
                },
                contributors: vec![],
                // this is fine, since if a link relation is used you should probably add a local
                // meta object anyways, in which case this code won't be reached.
                description: Some(format!("{:?}", &link)),
                icon_url: None,
                website_url: None,
            })
        },
        File::Maven {
            repository,
            artifact,
            ..
        } => {
            let repo = repos
                .get(&repository)
                .context("File references unknown repository!")?;

            match repo.repo_type {
                RepositoryType::Maven => Ok(Meta {
                    name: artifact,
                    contributors: vec![],
                    description: None,
                    icon_url: None,
                    website_url: None,
                }),
                RepositoryType::Curseforge => {
                    let (p_id, _) = util::parse_curseforge_artifact(&artifact)?;
                    let resp = http
                        .get(format!(
                            "https://addons-ecs.forgesvc.net/api/v2/addon/{}",
                            p_id
                        ))
                        .send()
                        .await?
                        .bytes()
                        .await?;

                    let resp = serde_json::from_slice::<AddonInfoResponse>(&resp)
                        .context("Failed to deserialize addon info!")?;

                    let icon_url = resp.attachments.into_iter().find_map(|a| {
                        if a.is_default {
                            Some(a.url.into())
                        } else {
                            None
                        }
                    });

                    Ok(Meta {
                        name: resp.name,
                        contributors: resp
                            .authors
                            .into_iter()
                            .map(|a| Contributor {
                                name: a.name,
                                roles: vec![],
                            })
                            .collect(),
                        description: Some(resp.summary),
                        icon_url,
                        website_url: Some(resp.website_url),
                    })
                },
            }
        },
    }
}

#[derive(Serialize)]
struct ModListContent {
    metas: Vec<Meta>,
    pack_meta: Meta,
}
