use std::path::PathBuf;
use structopt::StructOpt;

mod commands;
mod config;
mod downloader;
mod forge;
mod util;

#[derive(StructOpt)]
struct Opt {
    #[structopt(subcommand)]
    cmd: Command,
}

#[derive(StructOpt)]
enum Command {
    #[structopt(about = "Initializes a new modpack")]
    Init {
        #[structopt(help = "The name of the modpack")]
        modpack_name: String,
        #[structopt(help = "Name of the modpack author")]
        name: String,
        #[structopt(help = "The minecraft version of the modpack")]
        mcversion: String,
    },

    #[structopt(name = "downloadmods", about = "Downloads mods of the pack")]
    DownloadMods {
        #[structopt(help = "Directory to download mods to")]
        dir: PathBuf,
        #[structopt(short, long, help = "Download all relations and not just mods")]
        all: bool,
    },

    #[structopt(name = "buildtwitch", about = "Builds a twitch export of the pack")]
    BuildTwitch {
        #[structopt(
            short,
            long,
            help = "Downloads all relations instead of just required ones"
        )]
        all: bool,
    },

    #[structopt(about = "Deletes artifacts and temporary files")]
    Clean,

    #[structopt(
        name = "createmodlist",
        about = "Creates a HTML list of the pack's mods."
    )]
    CreateModList {
        #[structopt(
            default_value = "build/modlist.html",
            help = "File to write the mod list to"
        )]
        outfile: PathBuf,
    },
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let Opt { cmd } = Opt::from_args();

    match cmd {
        Command::Init {
            modpack_name,
            name,
            mcversion,
        } => commands::init::run(modpack_name, name, mcversion).await?,

        Command::DownloadMods { dir, all } => {
            commands::downloadmods::run(util::parse_config_and_manifest().await?, dir, all).await?
        },

        Command::BuildTwitch { all } => {
            commands::buildtwitch::run(util::parse_config_and_manifest().await?, all).await?
        },

        Command::Clean => commands::clean::run(util::parse_config().await?).await?,

        Command::CreateModList { outfile } => {
            commands::createmodlist::run(util::parse_config_and_manifest().await?, outfile).await?
        },
    }

    Ok(())
}
